/*-
 * Copyright (C)2005 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: codec.c,v 1.9 2005/11/29 01:55:33 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <babolo/BLINflag.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include "iso8583.h"
#ifndef ENOATTR
#define ENOATTR EINVAL
#endif
#define ZSIZE 10
#define UPHALF(A)   (((A) + 1) / 2)
#define BITUMASK(A) ((u_int64_t)1 << (64 - (A)))
#define TIMISU 0x8000
#define TIMSNL 0x0001
#define TIMDNL 0x0002
#define TIMNEQ 0x0004
/*
 * ISU NEQ DNL SNL
 *  0   0   0   0
 *  0   0   0   1 set unset
 *  0   0   1   0 unset set
 *  0   0   1   1
 *  0   1   0   0 set   set
 *  1   0   0   0
 */

/*    codex.len() */
#define MS_NUL 0
#define MS_MOD 1
#define MS_ILL 2 /*   */
#define MS_NOT 3 /*  */

static int
kolo(int mod, int hint, int cond) {
    if  ((hint < (mod >> 1)) && (cond > (hint + (mod >> 1)))) return (-1);
    if  ((hint > (mod >> 1)) && (cond < (hint - (mod >> 1)))) return (1);
    return(0);
}

#include "val.cc"
#include "ref.cc"
#include "asn.cc"
#include "cpy.cc"
#include "len.cc"

static const struct {
    ssize_t   (*len)(ISO8583_Pack *pack, int fno);
    u_int64_t (*val)(ISO8583_Pack *pack, int fno);
    char     *(*ref)(ISO8583_Pack *pack, int fno);
    int       (*asn)(ISO8583_Pack *pack, int fno, u_int64_t val);
    int       (*cpy)(ISO8583_Pack *pack, int fno, u_char *vly, ssize_t len);
} codex[] = {
 { eqb, ltrl, eqref, atrl, NULL} /* B_LITERAL  */
,{ eqa, NULL, eqref, NULL, cpyc} /* I_CHAR     */
,{  ll, NULL, llref, NULL, cpyl} /* B_LLCHAR   */
,{ lll, NULL,  NULL, NULL, NULL} /* B_LLLCHAR  */
,{ eqn,  lvr,  NULL,  avr, NULL} /* B_NUMERIC  */
,{ eqn,  lvl,  NULL,  avl, NULL} /* B_NUMERICl */
,{ lln, lvnr,  NULL, avnr, NULL} /* B_LLNUM    */
,{ lln, lvnl,  NULL, avnl, NULL} /* B_LLNUMl   */
,{ eqd,  tim,   dim,  aim, NULL} /* B_TIME     */
,{ eqd,  dat,   dad,  aat, NULL} /* B_DATE     */
,{ eqd, datt,  dadt, aatt, NULL} /* B_DATET    */
,{unkn, NULL,  NULL, NULL, NULL} /* B_BINARY   */
,{unkn, NULL,  NULL, NULL, NULL} /* B_BINARY   */
,{unkn, NULL,  NULL, NULL, NULL} /* B_BINARY   */
,{ eqd, timl,  diml, aiml, NULL} /* B_TIMEl    */
,{ eqd, datl,  dadl, aatl, NULL} /* B_DATEl    */
,{ eqd, dttl,  dddl, attl, NULL} /* B_DATETl   */
};

void
ISO8583_Save(ISO8583_Pack **pack) {
    free((*pack)->tz);
    free((*pack));
    *pack = NULL;
}

void
ISO8583_Free(ISO8583_Pack **pack) {
    if  ((*pack)->flags & ISO8583_FMS) free((*pack)->refo[0]);
    ISO8583_Save(pack);
}

int
ISO8583_Test(ISO8583_Pack *pack, int fno, u_int32_t mode) {
    int e = 0;

    if  (fno < 0) {
        ifBLIN_QV2(pack->flags) warnx("ISO8583_Test: Illegal index [%d]", fno);
        errno = EINVAL;
    } else {
        switch (mode & ISO8583_TEST) {
        case ISO8583_STAT:
            if  (fno < pack->package->bound && pack->refo[fno]) {
                e |= ISO8583_PRES;
                if  (codex[pack->package->fld[fno].convertor].val) e |= ISO8583_VALU;
                if  (codex[pack->package->fld[fno].convertor].ref) e |= ISO8583_REFU;
            }
            break;
        case ISO8583_SIZE:
            if  (  fno < pack->package->bound
                && pack->refo[fno]
                && (e = (codex[pack->package->fld[fno].convertor].len)(pack, fno)) <= 0
                ) {
                errno = EDOM;
                switch(-e) {
                case MS_NUL:
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Test: null member %d length\n", fno);
                    break;
                case MS_ILL:
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Test: invalid member %d length\n", fno);
                    break;
                default:
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Test: unverified member %d\n", fno);
                    errno = ENOSYS;
            }   }
            break;
        default:
            errno = EDOM;
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Test: invalid mode query\n", fno);
    }   }
    return(e);
}

static ISO8583_Pack *
mark(ISO8583_Pack **pack, int fno) {
    ssize_t l;
    if  ((*pack)->package->fld[fno].convertor >= ISO8583_E_MAX) {
        ifBLIN_QV2((*pack)->flags) warnx("ISO8583_Mark: invalid package convertor %d\n", fno);
        errno = EINVAL;
        ISO8583_Free(pack);
    } else {
        if  ((l = ISO8583_Test(*pack, fno, ISO8583_SIZE)) <= 0) {
            ISO8583_Free(pack);
        } else if ((*pack)->msz && (*pack)->totlen + l > (*pack)->msz) {
            ifBLIN_QV2((*pack)->flags)
                warnx("ISO8583_Mark: size overload %u [%d]\n", (*pack)->msz, fno);
            errno = EOVERFLOW;
            ISO8583_Free(pack);
        } else {
            (*pack)->totlen += l;
    }   }
    return(*pack);
}

ISO8583_Pack *
ISO8583_Mark(u_int32_t flags, const ISO8583_Package *package, u_char *in, size_t msz, char *tzname) {
    int i;
    u_int64_t m;
    ISO8583_Pack *pack = NULL;

    if  (!package) {
        ifBLIN_QV2(flags) warnx("ISO8583_Mark: package absent\n");
        errno = EINVAL;
    } else {
        pack = calloc(sizeof(ISO8583_Pack), 1);
        if  (!pack) {
            ifBLIN_QV2(flags) warnx("ISO8583_Mark: calloc pack failed\n");
            errno = ENOMEM;
        } else {
            pack->flags = flags & ~ISO8583_FMS;
            pack->package = package;
            if  ((pack->tz = tzname) && !(pack->tz = strdup(tzname))) {
                ifBLIN_QV2(flags) warnx("ISO8583_Mark: strdup failed\n");
                errno = ENOMEM;
                ISO8583_Free(&pack);
            } else if (!(flags & ISO8583_LTM) && (time(&(pack->now)) < 0)) {
                ifBLIN_QV2(flags) warn("ISO8583_Mark: time failed");
                ISO8583_Free(&pack);
            } else if (in) {
                pack->msz = msz;
            ;   pack->totlen = 0;
            ;   pack->refo[0] = in;
            ;   if  (mark(&pack, 0)) {
            ;       pack->refo[1] = pack->refo[0] + pack->totlen;
            ;       m = ISO8583_Val(pack, 1) | pack->package->premask;
            ;       ifBLIN_QV3(flags) fprintf(stderr, "m=%08llX\n", m);
            ;       for (i = 1; m; i++, m <<= 1) {
            ;           if  (m & 0x8000000000000000L) {
            ;               pack->refo[i] = pack->refo[0] + pack->totlen;
            ;               if  (!mark(&pack, i)) break;
                }   }   }
            } else {
                if  (!msz) msz = ((flags & ISO8583_MSZ) == 0) ? PAGE_SIZE : (flags & ISO8583_MSZ);
            ;   pack->msz = msz;
            ;   pack->totlen = 0;
            ;   pack->flags |= ISO8583_FMS;
            ;   pack->refo[0] = calloc(msz, 1);
            ;   if  (!pack->refo[0]) {
            ;       ifBLIN_QV2(flags) warnx("ISO8583_Mark: calloc pack->refo[0] failed\n");
            ;       errno = ENOMEM;
            ;       ISO8583_Free(&pack);
    }   }   }   }
    return(pack);
}

u_int64_t
ISO8583_Val(ISO8583_Pack *pack, int fno) {
    u_int64_t val = 0;

    if  (fno < 0 || fno >= pack->package->bound) {
        ifBLIN_QV2(pack->flags) warnx("ISO8583_Val: Illegal index [%d]", fno);
        errno = EINVAL;
    } else if (  (ISO8583_Test(pack, fno, ISO8583_STAT) & (ISO8583_PRES | ISO8583_VALU))
              == (ISO8583_PRES | ISO8583_VALU)
              ) {
        val = (codex[pack->package->fld[fno].convertor].val)(pack, fno);
    }
    return(val);
}

char *
ISO8583_Ref(ISO8583_Pack *pack, int fno) {
    char *ref = NULL;

    if  (fno < 0 || fno >= pack->package->bound) {
        ifBLIN_QV2(pack->flags) warnx("ISO8583_Ref: Illegal index [%d]", fno);
        errno = EINVAL;
    } else if (  (ISO8583_Test(pack, fno, ISO8583_STAT) & (ISO8583_PRES | ISO8583_REFU))
              == (ISO8583_PRES | ISO8583_REFU)
              ) {
        ref = (codex[pack->package->fld[fno].convertor].ref)(pack, fno);
    }
    return(ref);
}

int
ISO8583_Asn(ISO8583_Pack *pack, int fno, u_int64_t val) {
    u_char *ref;
    ssize_t e = 0;
    int (*asn)(ISO8583_Pack *pack, int fno, u_int64_t val);
    int psft;

    ifBLIN_QV4(pack->flags) fprintf( stderr, "+ISO8583_Asn [%2d]=%llu %016llX\n"
                                   , fno, val
                                   , BITUMASK(fno - 1) - 1
                                   );
    if  (fno < 0 || fno >= pack->package->bound) {
        ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal index [%d]", fno);
        errno = EINVAL;
    } else {
        ref = pack->refo[fno];
        asn = codex[pack->package->fld[fno].convertor].asn;
        psft = pack->package->flags & ISO8583_PSFT;
        if  (!asn) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: No Asn");
            errno = EFAULT;
        } else if ((fno == 0) && (fno < psft)) {
            if  (!ref) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal ISO8583_Pack");
                errno = EFAULT;
            } else if (!(e = (*asn)(pack, fno, val))) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
                errno = EINVAL;
            } else if (!pack->totlen) {
                pack->totlen += e;
            }
        } else if ((fno != 0) && (fno < psft)) {
            int n = 1;
            if  (!ref) {
                for (n = fno; n < pack->package->bound; n++) {
                    if  (pack->refo[n]) {
                        n = 0;
                        break;
            }   }   }
            if  (!n) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal sequence [%d]", fno);
                errno = EINVAL;
            } else {
                n = 0;
                if  (!ref) {
                    pack->refo[fno] = pack->refo[0] + pack->totlen;
                    n++;
                }
                if  (!(e = (*asn)(pack, fno, val))) {
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
                    errno = EINVAL;
                } else if (n) {
                    pack->totlen += e;
            }   }
        } else if (fno == psft) {
            if  (ref) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal sequence [%d]", fno);
                errno = EINVAL;
            } else if (val) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal initiation [%d]=%llu", fno, val);
                errno = EINVAL;
            } else {
                pack->refo[fno] = pack->refo[0] + pack->totlen;
                if  (!(e = (*asn)(pack, fno, val))) {
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
                    errno = EINVAL;
                } else {
                    pack->totlen += e;
            }   }
        } else if (!ref && (BITUMASK(fno - 1) - 1) & ISO8583_Val(pack, psft)) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal sequence [%d]", fno);
            errno = EINVAL;
        } else {
            int n = 0;
            if  (!ref) {
                pack->refo[fno] = pack->refo[0] + pack->totlen;
                n++;
            }
            if  (!(e = (*asn)(pack, fno, val))) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: Illegal value [%d]=%llu", fno, val);
                errno = EINVAL;
            } else if (n) {
                pack->totlen += e;
                if  (!atrl(pack, psft, (ISO8583_Val(pack, psft) | BITUMASK(fno)))) {
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Asn: could not set bitmap");
                    errno = ENOATTR;
    }   }   }   }
    ifBLIN_QV4(pack->flags) fprintf(stderr, "-ISO8583_Asn %d totlen=%u\n", e, pack->totlen);
    return(e);
}

int
ISO8583_Cpy(ISO8583_Pack *pack, int fno, u_char *vly, ssize_t len) {
    u_char *ref;
    ssize_t e = 0, l = len;
    int (*cpy)(ISO8583_Pack *pack, int fno, u_char *vly, ssize_t len);
    int psft;

    ifBLIN_QV4(pack->flags) fprintf( stderr, "+ISO8583_Cpy [%2d] %016llX\n"
                                   , fno
                                   , BITUMASK(fno - 1) - 1
                                   );
    if  (!len) l = strlen(vly);
    ifBLIN_QV4(pack->flags) fprintf( stderr, "len=%d\n", l);
    if  (fno < 0 || fno >= pack->package->bound) {
        ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: Illegal index [%d]", fno);
        errno = EINVAL;
    } else {
        ref = pack->refo[fno];
        cpy = codex[pack->package->fld[fno].convertor].cpy;
        psft = pack->package->flags & ISO8583_PSFT;
        if  (!cpy) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: No Cpy");
            errno = EFAULT;
        } else if (fno <= psft) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: Not implemented [%d]", fno);
            errno = ENOSYS;
        } else if (!ref && (BITUMASK(fno - 1) - 1) & ISO8583_Val(pack, psft)) {
            ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: Illegal sequence [%d]", fno);
            errno = EINVAL;
        } else {
            int n = 0;
            if  (!ref) {
                pack->refo[fno] = pack->refo[0] + pack->totlen;
                n++;
            }
            if  (!(e = (cpy)(pack, fno, vly, l))) {
                ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: Illegal copy [%d]", fno);
                errno = EINVAL;
            } else if (n) {
                pack->totlen += e;
                if  (!atrl(pack, psft, (ISO8583_Val(pack, psft) | (BITUMASK(fno))))) {
                    ifBLIN_QV2(pack->flags) warnx("ISO8583_Cpy: could not set bitmap");
                    errno = ENOATTR;
    }   }   }   }
    ifBLIN_QV4(pack->flags) fprintf(stderr, "-ISO8583_Cpy %d totlen=%u\n", e, pack->totlen);
    return(e);
}

int
ISO8583_Control(u_int32_t flags, const ISO8583_Package *package) {
    int e = 0;
    if  (  package->flags != 1
        || package->fld[0].convertor != ISO8583_B_NUMERIC
        || package->fld[1].convertor != ISO8583_B_LITERAL
        || package->fld[0].len != 4
        || package->fld[1].len != 8
        ) {
        e = -1;
    }
    return(e);
}
