/*-
 * Copyright (C)2005 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: iso8583.c,v 1.6 2005/11/27 01:19:26 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <fcntl.h>
#include <mife.h>
#include <time.h>
#include <err.h>
#include "../iso8583.h"
#define ISO8583_LOMSB 0x0100000
#define ISO8583_LOINC 0x0200000

static void
dmps(u_int32_t a, int s, u_char *c, ssize_t l, ssize_t m) {
    int d;
    printf("%04X(%2d)", a, s);
    if  (m > 0) {
        printf(" ");
        for (d = 0; d < l && d < m; d++) printf("%02X", c[d]);
        for (d = l; d < m; d++) printf("  ");
}   }

static void
usage(int ex) {
    exit(ex);
}

int
main(int argc, char *argv[], char **envp) {
    const ISO8583_Package *thispck;
    struct mife_descriptor *in;
    ISO8583_Pack *pack;
    u_int32_t flags, lofl = 0;
    ssize_t packlen, ofl = 0;
    u_char *inp, *r;
    int c, infl, dmp = 0;
    char *tz = NULL;
    struct tm *tm;
    time_t clock, locat = 0;

    flags = BLIN_VER1;
    infl = fileno(stdin);
    if  (time(&clock) == -1) {
        ifBLIN_QV1(flags) warnx("time() failed");
        exit(EX_OSERR);
    }
    while ((c = getopt(argc, argv, "d:if:l:nqt:v")) != EOF) {
        switch (c) {
        case 'd':
            dmp = strtol(optarg, NULL, 0);
            break;
        case 'i':
            lofl |= ISO8583_LOINC;
            break;
        case 'f':
            /*  if  (infl <> fileno(stdin)) close(infl); */
            if  ((infl = open(optarg, O_RDONLY)) == NULL) {
                ifBLIN_QV1(flags) warn("input %s failed", optarg);
                exit(EX_NOINPUT);
            }
            break;
        case 'l':
            if  (time(&clock) == -1) {
                ifBLIN_QV1(flags) warnx("time() failed");
                exit(EX_OSERR);
            } else if (!(strptime(optarg, "%F %T", (tm = gmtime(&clock))))) {
                ifBLIN_QV1(flags) warnx("Time %s failed", optarg);
                exit(EX_DATAERR);
            }
            locat = timegm(tm);
            flags |= ISO8583_LTM;
        case 'n':
            lofl |= ISO8583_LOMSB;
            break;
        case 'q':
            flags &= ~BLIN_VERMASK;
            break;
        case 't':
            tz = optarg;
            break;
        case 'v':
            BLIN_VERBOSE(flags);
            break;
        case '?':
            usage(EX_OK);
        default:
            usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc == 0) exit(EX_OK);
    if  (!strcmp(*argv, "list")) {
        ISO8583_List(0, 0, stdout);
    } else {
        if  (argc < 2) {
            ifBLIN_QV1(flags) warnx("command and format number needed");
            exit(EX_USAGE);
        }
        if  (!(thispck = ISO8583_Take(argv[1]))) {
            ifBLIN_QV1(flags) warnx("No message format");
            exit(EX_NOINPUT);
        } else if  (!strcmp(*argv, "get")) {
            if  (argc > 2) ofl = strtoul(argv[2], NULL, 0);
        ;   if  ((in = mife_opef(0, infl)) == NULL) {
        ;       ifBLIN_QV1(flags) warn("input file failed");
        ;       exit(EX_NOINPUT);
        ;   }
        ;   if  ((packlen = mife_read(in, 0, 0, NULL)) < 0) {
        ;           ifBLIN_QV1(flags) warn("input read failed");
        ;           exit(EX_NOINPUT);
        ;   }
        ;   if  (!(inp = mife_get(in, 0))) {
        ;           ifBLIN_QV1(flags) warn("input get failed");
        ;           exit(EX_NOINPUT);
        ;   }
        ;   if  (!(pack = ISO8583_Mark(flags, thispck, inp + ofl, 0, tz))) {
        ;           ifBLIN_QV1(flags) warn("message not marked");
        ;           exit(EX_SOFTWARE);
        ;   }
        ;   if  (flags & ISO8583_LTM) pack->now = locat;
        ;   ifBLIN_QV3(flags) fprintf(stderr, "now=%lu\n", pack->now);
        ;   ifBLIN_QV2(flags) {
                ssize_t cl;
        ;       cl = 0;
        ;       if  (lofl & ISO8583_LOMSB) {
        ;           for (c = 0; c < ofl; c++) cl = (cl << 8) | (inp[c] & 0xFF);
        ;       } else {
        ;           for (c = ofl; c > 0;) cl = (cl << 8) | (inp[--c] & 0xFF);
        ;       }
        ;       printf("     ");
        ;       dmps(0, ofl, inp, ofl, dmp);
        ;       printf(" length received=%u claimed=%u, marked=%u\n", packlen, cl, pack->totlen);
        ;   }
        ;   for (c = 0; c < pack->package->bound; c++) {
        ;       if  (pack->refo[c]) {
        ;           printf("[%2d]=", c);
        ;           dmps( (u_int32_t)(pack->refo[c] - inp)
                        , ISO8583_Test(pack, c, ISO8583_SIZE)
                        , pack->refo[c]
                        , ISO8583_Test(pack, c, ISO8583_SIZE)
                        , dmp
                        );
        ;           printf(" Test=%04X", ISO8583_Test(pack, c, ISO8583_STAT));
        ;           if  (  (ISO8583_Test(pack, c, ISO8583_STAT) & (ISO8583_PRES | ISO8583_VALU))
                        == (ISO8583_PRES | ISO8583_VALU)
                        ) {
        ;               printf(" Val=%llu", ISO8583_Val(pack, c));
        ;           }
        ;           r = ISO8583_Ref(pack, c);
        ;           if  (r) printf(" Ref=\"%s\"", r);
        ;           printf("  %s\n", thispck->fld[c].description);
        ;           free(r);
        ;   }   }
        ;   printf(" end=%04X\n", (u_int32_t)(pack->totlen + ofl));
            ISO8583_Free(&pack);
        } else if (!strcmp(*argv, "put")) {
            struct babolo_parm *bp;
            u_int64_t val;
            int fno;
        ;   if  (argc > 2) ofl = strtoul(argv[2], NULL, 0);
        ;   if  (ofl > 4) {
        ;       ifBLIN_QV1(flags) warn("Invalid offset %d", ofl);
        ;       exit(EX_DATAERR);
        ;   }
        ;   if  ((in = mife_opef(MIFE_FULL, infl)) == NULL) {
        ;       ifBLIN_QV1(flags) warn("input file failed");
        ;       exit(EX_NOINPUT);
        ;   }
        ;   if  ((packlen = mife_read(in, 0, 0, NULL)) < 0) {
        ;           ifBLIN_QV1(flags) warn("input read failed");
        ;           exit(EX_NOINPUT);
        ;   }
        ;   if  (!(r = inp = mife_get(in, 0))) {
        ;           ifBLIN_QV1(flags) warn("input get failed");
        ;           exit(EX_NOINPUT);
        ;   }
        ;   if  (!(pack = ISO8583_Mark(flags, thispck, NULL, 0, tz))) {
        ;           ifBLIN_QV1(flags) warn("message not created");
        ;           exit(EX_SOFTWARE);
        ;   }
        ;   for (c = 0; r < inp + packlen; c++) {
        ;       bp = babolo_getparm(Bpars_VTLN | Bpars_CRLN | Bpars_LFLN | kukBpars_NPRM(3), &r, "");
        ;       if  (!bp) {
        ;           ifBLIN_QV1(flags) warn("Input string error \"%s\"", r);
        ;           exit(EX_NOINPUT);
        ;       } else if (bp->argc > 2) {
        ;           fno = strtol(bp->argv[0], NULL, 0);
        ;           if  (!strcmp(bp->argv[1], "num")) {
        ;               val = strtoull(bp->argv[2], NULL, 0);
        ;               if  (!(ISO8583_Asn(pack, fno, val))) {
        ;                   ifBLIN_QV1(flags) warnx("Assign [%d]=%llu failed", fno, val);
        ;                   exit(EX_DATAERR);
        ;               }
        ;           } else if (!strcmp(bp->argv[1], "str")) {
        ;               if  (!(ISO8583_Cpy(pack, fno, bp->argv[2], 0))) {
        ;                   ifBLIN_QV1(flags) warnx("Copy [%d] failed", fno);
        ;                   exit(EX_DATAERR);
        ;               }
        ;           } else if (!strcmp(bp->argv[1], "tim")) {
        ;               if  (time(&clock) == -1) {
        ;                   ifBLIN_QV1(flags) warnx("time() failed");
        ;                   exit(EX_OSERR);
        ;               } else if (!(strptime(bp->argv[2], "%F %T", (tm = gmtime(&clock))))) {
        ;                   ifBLIN_QV1(flags) warnx("Time [%d]=%s failed", fno, bp->argv[2]);
        ;                   exit(EX_DATAERR);
        ;               } else if (!(ISO8583_Asn(pack, fno, timegm(tm)))) {
        ;                   ifBLIN_QV1(flags) warnx("Assign [%d]=time failed", fno);
        ;                   exit(EX_DATAERR);
        ;               }
        ;           } else {
        ;               ifBLIN_QV1(flags) warnx("No conversion %s in [%d]", bp->argv[1], fno);
        ;               exit(EX_DATAERR);
        ;           }
        ;       } else if (bp->argc > 0) {
        ;           fno = strtol(bp->argv[0], NULL, 0);
        ;           ifBLIN_QV1(flags) warnx("Assign [%d] emty", fno);
        ;           exit(EX_DATAERR);
        ;       }
        ;       babolo_freeparm(bp);
        ;   }
        ;   val = pack->totlen;
        ;   if  (lofl & ISO8583_LOINC) val += ofl;
        ;   if  (lofl & ISO8583_LOMSB) val = htonl((u_long)val);
        ;   if  (mife_writ( fileno(stdout)
                          , &((u_char*)(&val))[(lofl & ISO8583_LOMSB) ? sizeof(u_long) - ofl : 0]
                          , ofl
                ) < 0     ) {
        ;       ifBLIN_QV1(flags) warn("output write failed");
        ;       exit(EX_IOERR);
        ;   }
        ;   if  (mife_writ(fileno(stdout), pack->refo[0], pack->totlen) < 0) {
        ;       ifBLIN_QV1(flags) warn("output write failed");
                exit(EX_IOERR);
    }   }   }
    exit(EX_OK);
}
